import json
import os
import sys

import bpy #pylint: disable=import-error


def export_gltf(settings, src, dst):
    print('Converting .blend file ({}) to .gltf ({})'.format(src, dst))

    # Lazy-load blendergltf
    scriptdir = os.path.dirname(__file__)
    sys.path.insert(0, os.path.join(scriptdir, 'blendergltf'))
    sys.path.insert(0, os.path.join(scriptdir))
    if 'blendergltf' in sys.modules:
        del sys.modules['blendergltf']
    from gltfexts import ExtMaterialsLegacy, ExtZup #pylint: disable=import-error
    import blendergltf #pylint: disable=import-error

    available_extensions = blendergltf.extensions

    dstdir = os.path.dirname(dst)
    os.makedirs(dstdir, exist_ok=True)

    gltf_settings = {
        'extension_exporters': [
            available_extensions.khr_lights.KhrLights(),
            available_extensions.blender_physics.BlenderPhysics(),
            ExtZup(),
        ],
        'gltf_output_dir': dstdir,
        'images_data_storage': 'REFERENCE',
        'nodes_export_hidden': True,
        'meshes_interleave_vertex_data': False,
    }

    if settings['material_mode'] == 'legacy':
        gltf_settings['extension_exporters'].append(ExtMaterialsLegacy())

    collections = [
        "actions",
        "cameras",
        "images",
        "lamps",
        "materials",
        "meshes",
        "objects",
        "scenes",
        "textures",
    ]
    scene = {
        cname: list(getattr(bpy.data, cname))
        for cname in collections
    }
    gltfdata = blendergltf.blendergltf.export_gltf(scene, gltf_settings)

    with open(dst, 'w') as gltffile:
        json.dump(gltfdata, gltffile, sort_keys=True, indent=4)


def main():
    args = sys.argv[sys.argv.index('--')+1:]

    #print(args)
    settings_fname, srcroot, dstdir, blendfiles = args[0], args[1], args[2], args[3:]

    print('srcroot:', srcroot)
    print('Exporting:', blendfiles)
    print('Export to:', dstdir)

    with open(settings_fname) as settings_file:
        settings = json.load(settings_file)

    for blendfile in blendfiles:
        src = blendfile
        dst = src.replace(srcroot, dstdir).replace('.blend', '.gltf')

        bpy.ops.wm.open_mainfile(filepath=src)
        export_gltf(settings, src, dst)


if __name__ == '__main__':
    main()



from .base import BaseExporter


class CameraExporter(BaseExporter):
    gltf_key = 'cameras'
    blender_key = 'cameras'

    @classmethod
    def export(cls, state, blender_data):
        camera_gltf = {}
        if blender_data.type == 'ORTHO':
            xmag = 0.5 * blender_data.ortho_scale
            ymag = xmag * state['aspect_ratio']
            camera_gltf = {
                'orthographic': {
                    'xmag': ymag,
                    'ymag': xmag,
                    'zfar': blender_data.clip_end,
                    'znear': blender_data.clip_start,
                },
                'type': 'orthographic',
            }
        else:
            angle_y = blender_data.angle_y if blender_data.angle_y != 0.0 else 1e-6
            camera_gltf = {
                'perspective': {
                    'aspectRatio': blender_data.angle_x / angle_y,
                    'yfov': angle_y,
                    'zfar': blender_data.clip_end,
                    'znear': blender_data.clip_start,
                },
                'type': 'perspective',
            }
        camera_gltf['name'] = blender_data.name
        extras = cls.get_custom_properties(blender_data)
        if extras:
            camera_gltf['extras'] = extras
        return camera_gltf



from ..blendergltf import Reference


class BlenderPhysics:
    ext_meta = {
        'name': 'BLENDER_physics',
        'url': (
            'https://github.com/Kupoman/blendergltf/tree/master/extensions/'
            'BLENDER_physics'
        ),
        'isDraft': True,
    }

    def export_physics(self, state, obj, gltf_node):
        body = obj.rigid_body
        bounds = [obj.dimensions[i] / gltf_node['scale'][i] for i in range(3)]
        collision_layers = sum(layer << i for i, layer in enumerate(body.collision_groups))
        physics = {
            'collisionShapes': [{
                'shapeType': body.collision_shape.upper(),
                'boundingBox': bounds,
                'primaryAxis': "Z",
            }],
            'mass': body.mass,
            'static': body.type == 'PASSIVE',
            'collisionGroups': collision_layers,
            'collisionMasks': collision_layers,
        }

        if body.collision_shape in ('CONVEX_HULL', 'MESH'):
            mesh = state['mod_meshes'].get(obj.name, obj.data)
            meshref = Reference('meshes', mesh.name, physics['collisionShapes'][0], 'mesh')
            physics['collisionShapes'][0]['mesh'] = meshref
            state['references'].append(meshref)

        return physics

    def export(self, state):
        state['extensions_used'].append('BLENDER_physics')

        obj_pairs = [
            (obj, state['output']['nodes'][state['refmap'][('objects', obj.name)]])
            for obj in state['input']['objects']
            if getattr(obj, 'rigid_body', False)
        ]
        for obj, node in obj_pairs:
            node['extensions'] = node.get('extensions', {})
            node['extensions']['BLENDER_physics'] = self.export_physics(state, obj, node)

        scene_pairs = [
            (scene, state['output']['scenes'][state['refmap'][('scenes', scene.name)]])
            for scene in state['input']['scenes']
        ]
        for scene, gltf in scene_pairs:
            gltf['extensions'] = gltf.get('extensions', {})
            gltf['extensions']['BLENDER_physics'] = {
                'gravity': scene.gravity.z
            }



from distutils.version import StrictVersion as Version

from ..blendergltf import Reference
from ..exporters import BaseExporter


def export_light(light):
    def calc_att():
        constant_factor = 1
        linear_factor = 0
        quad_factor = 0

        if light.falloff_type == 'INVERSE_LINEAR':
            linear_factor = 1 / light.distance
        elif light.falloff_type == 'INVERSE_SQUARE':
            quad_factor = 1 / light.distance
        elif light.falloff_type == 'LINEAR_QUADRATIC_WEIGHTED':
            linear_factor = light.linear_attenuation * (1 / light.distance)
            quad_factor = light.quadratic_attenuation * (1 / (light.distance * light.distance))
        elif light.falloff_type == 'INVERSE_COEFFICIENTS':
            constant_factor = light.constant_coefficient
            linear_factor = light.linear_coefficient * (1 / light.distance)
            quad_factor = light.quadratic_coefficient * (1 / (light.distance * light.distance))

        return constant_factor, linear_factor, quad_factor

    gltf_light = {}
    if light.type == 'SUN':
        gltf_light = {
            'directional': {
                'color': (light.color * light.energy)[:],
            },
            'type': 'directional',
        }
    elif light.type == 'POINT':
        constant_factor, linear_factor, quad_factor = calc_att()
        gltf_light = {
            'point': {
                'color': (light.color * light.energy)[:],

                'constantAttenuation': constant_factor,
                'linearAttenuation': linear_factor,
                'quadraticAttenuation': quad_factor,
            },
            'type': 'point',
        }
    elif light.type == 'SPOT':
        constant_factor, linear_factor, quad_factor = calc_att()
        gltf_light = {
            'spot': {
                'color': (light.color * light.energy)[:],

                'fallOffAngle': light.spot_size,
                'fallOffExponent': 128.0 * light.spot_blend,
                'constantAttenuation': constant_factor,
                'linearAttenuation': linear_factor,
                'quadraticAttenuation': quad_factor,
            },
            'type': 'spot',
        }
    else:
        print("Unsupported lamp type on {}: {}".format(light.name, light.type))
        gltf_light = {'type': 'unsupported'}

    gltf_light['name'] = light.name

    extras = BaseExporter.get_custom_properties(light)
    if extras:
        gltf_light['extras'] = extras

    return gltf_light


def export_lights(state, extension_name):
    # Export lights
    state['output']['extensions'] = state['output'].get('extensions', {})
    state['output']['extensions'][extension_name] = {
        'lights': [export_light(lamp) for lamp in state['input'].get('lamps', [])]
    }

    # Add light references to nodes
    obj_pairs = [
        (obj, state['output']['nodes'][state['refmap'][('objects', obj.name)]])
        for obj in state['input']['objects']
    ]

    for obj, node in obj_pairs:
        if getattr(obj, 'type', '') == 'LAMP':
            node['extensions'] = node.get('extensions', {})
            ext = node['extensions'][extension_name] = {}
            ext['light'] = Reference('lamps', obj.data.name, ext, 'light')
            state['references'].append(ext['light'])

    # Convert light list to dictionary for glTF 1.0
    if state['version'] < Version('2.0'):
        extension = state['output']['extensions'][extension_name]
        extension['lights'] = {
            'lights_' + light['name']: light for light in extension['lights']
        }



from distutils.version import StrictVersion as Version

import bpy

from .base import BaseExporter
from .common import (
    Reference,
    SimpleID,
)


# Texture formats
GL_ALPHA = 6406
GL_RGB = 6407
GL_RGBA = 6408
GL_LUMINANCE = 6409
GL_LUMINANCE_ALPHA = 6410

# Texture filtering
GL_NEAREST = 9728
GL_LINEAR = 9729
GL_LINEAR_MIPMAP_LINEAR = 9987

# Texture wrapping
GL_CLAMP_TO_EDGE = 33071
GL_MIRRORED_REPEAT = 33648
GL_REPEAT = 10497

# sRGB texture formats (not actually part of WebGL 1.0 or glTF 1.0)
GL_SRGB = 0x8C40
GL_SRGB_ALPHA = 0x8C42


class TextureExporter(BaseExporter):
    gltf_key = 'textures'
    blender_key = 'textures'

    @classmethod
    def check(cls, state, blender_data):
        errors = []
        if not isinstance(blender_data, bpy.types.ImageTexture):
            errors.append('is not an ImageTexture')
        elif blender_data.image is None:
            errors.append('has no image reference')
        elif blender_data.image.channels not in [3, 4]:
            errors.append(
                'points to {}-channel image (must be 3 or 4)'
                .format(blender_data.image.channels)
            )

        if errors:
            err_list = '\n\t'.join(errors)
            print(
                'Unable to export texture {} due to the following errors:\n\t{}'
                .format(blender_data.name, err_list)
            )
            return False

        return True

    @classmethod
    def export_sampler(cls, blender_data):
        gltf_sampler = {
            'name': blender_data.name,
        }

        # Handle wrapS and wrapT
        if blender_data.extension in ('REPEAT', 'CHECKER', 'EXTEND'):
            if blender_data.use_mirror_x:
                gltf_sampler['wrapS'] = GL_MIRRORED_REPEAT
            else:
                gltf_sampler['wrapS'] = GL_REPEAT

            if blender_data.use_mirror_y:
                gltf_sampler['wrapT'] = GL_MIRRORED_REPEAT
            else:
                gltf_sampler['wrapT'] = GL_REPEAT
        elif blender_data.extension in ('CLIP', 'CLIP_CUBE'):
            gltf_sampler['wrapS'] = GL_CLAMP_TO_EDGE
            gltf_sampler['wrapT'] = GL_CLAMP_TO_EDGE
        else:
            print('Warning: Unknown texture extension option:', blender_data.extension)

        # Handle minFilter and magFilter
        if blender_data.use_mipmap:
            gltf_sampler['minFilter'] = GL_LINEAR_MIPMAP_LINEAR
            gltf_sampler['magFilter'] = GL_LINEAR
        else:
            gltf_sampler['minFilter'] = GL_NEAREST
            gltf_sampler['magFilter'] = GL_NEAREST

        return gltf_sampler

    @classmethod
    def export(cls, state, blender_data):
        gltf_texture = {
            'name': blender_data.name,
        }

        gltf_sampler = cls.export_sampler(blender_data)
        state['input']['samplers'].append(SimpleID(blender_data.name))
        state['samplers'].append(gltf_sampler)

        gltf_texture['sampler'] = Reference('samplers', blender_data.name, gltf_texture, 'sampler')
        state['references'].append(gltf_texture['sampler'])

        gltf_texture['source'] = Reference(
            'images',
            blender_data.image.name,
            gltf_texture,
            'source'
        )
        state['references'].append(gltf_texture['source'])

        if state['version'] < Version('2.0'):
            tformat = None
            channels = blender_data.image.channels
            image_is_srgb = blender_data.image.colorspace_settings.name == 'sRGB'
            use_srgb = state['settings']['images_allow_srgb'] and image_is_srgb

            if channels == 3:
                if use_srgb:
                    tformat = GL_SRGB
                else:
                    tformat = GL_RGB
            elif channels == 4:
                if use_srgb:
                    tformat = GL_SRGB_ALPHA
                else:
                    tformat = GL_RGBA

            gltf_texture['format'] = gltf_texture['internalFormat'] = tformat

        return gltf_texture



class KhrMaterialsUnlit:
    ext_meta = {
        'name': 'KHR_materials_unlit',
        'url': (
            'https://github.com/KhronosGroup/glTF/pull/1163'
        ),
        'isDraft': True,
    }

    def export(self, state):
        state['extensions_used'].append('KHR_materials_unlit')

        # Export materials
        material_pairs = [
            (material, state['output']['materials'][state['refmap'][('materials', material.name)]])
            for material in state['input']['materials']
        ]
        for bl_mat, gl_mat in material_pairs:
            if bl_mat.use_shadeless:
                print('Found shadeless material')
            gl_mat['extensions'] = gl_mat.get('extensions', {})
            gl_mat['extensions']['KHR_materials_unlit'] = {}



import base64
from distutils.version import StrictVersion as Version
import collections
import json
import math
import os
import struct

import bpy


def get_bone_name(bone):
    return '{}_{}'.format(bone.id_data.name, bone.name)


class AnimationPair:
    def __init__(self, target, action, is_shape_key=False):
        self.target = target
        self.action = action
        self.name = '{}_{}'.format(target.name, action.name)
        self.is_shape_key = is_shape_key


class Reference:
    __slots__ = (
        "blender_type",
        "blender_name",
        "source",
        "prop",
        "value",
    )

    def __init__(self, blender_type, blender_name, source, prop):
        self.blender_type = blender_type
        self.blender_name = blender_name
        self.source = source
        self.prop = prop
        self.value = None

    def __str__(self):
        return json.dumps([self.blender_type, self.blender_name, self.value])


class SimpleID:
    def __init__(self, name, data=None):
        self.name = name
        self.data = data


class Buffer:
    ARRAY_BUFFER = 34962
    ELEMENT_ARRAY_BUFFER = 34963

    BYTE = 5120
    UNSIGNED_BYTE = 5121
    SHORT = 5122
    UNSIGNED_SHORT = 5123
    INT = 5124
    UNSIGNED_INT = 5125

    FLOAT = 5126

    MAT4 = 'MAT4'
    VEC4 = 'VEC4'
    VEC3 = 'VEC3'
    VEC2 = 'VEC2'
    SCALAR = 'SCALAR'

    class Accessor:
        __slots__ = (
            "name",
            "buffer",
            "buffer_view",
            "byte_offset",
            "byte_stride",
            "component_type",
            "count",
            "min",
            "max",
            "data_type",
            "type_size",
            "_ctype",
            "_ctype_size",
            "_buffer_data",
            )

        def __init__(self,
                     name,
                     buffer,
                     buffer_view,
                     byte_offset,
                     byte_stride,
                     component_type,
                     count,
                     data_type):
            self.name = name
            self.buffer = buffer
            self.buffer_view = buffer_view
            self.byte_offset = byte_offset
            self.byte_stride = byte_stride
            self.component_type = component_type
            self.count = count
            self.min = [math.inf for i in range(16)]
            self.max = [-math.inf for i in range(16)]
            self.data_type = data_type

            if self.data_type == Buffer.MAT4:
                self.type_size = 16
            elif self.data_type == Buffer.VEC4:
                self.type_size = 4
            elif self.data_type == Buffer.VEC3:
                self.type_size = 3
            elif self.data_type == Buffer.VEC2:
                self.type_size = 2
            else:
                self.type_size = 1

            if component_type == Buffer.BYTE:
                self._ctype = '<b'
            elif component_type == Buffer.UNSIGNED_BYTE:
                self._ctype = '<B'
            elif component_type == Buffer.SHORT:
                self._ctype = '<h'
            elif component_type == Buffer.UNSIGNED_SHORT:
                self._ctype = '<H'
            elif component_type == Buffer.INT:
                self._ctype = '<i'
            elif component_type == Buffer.UNSIGNED_INT:
                self._ctype = '<I'
            elif component_type == Buffer.FLOAT:
                self._ctype = '<f'
            else:
                raise ValueError("Bad component type")

            self._ctype_size = struct.calcsize(self._ctype)
            self._buffer_data = self.buffer.get_buffer_data(self.buffer_view)

        def __len__(self):
            return self.count

        def __getitem__(self, idx):
            if not isinstance(idx, int):
                raise TypeError("Expected an integer index")

            ptr = (
                (
                    (idx % self.type_size)
                    * self._ctype_size + idx // self.type_size * self.byte_stride
                ) + self.byte_offset
            )

            return struct.unpack_from(self._ctype, self._buffer_data, ptr)[0]

        def __setitem__(self, idx, value):
            if not isinstance(idx, int):
                raise TypeError("Expected an integer index")

            i = idx % self.type_size
            self.min[i] = value if value < self.min[i] else self.min[i]
            self.max[i] = value if value > self.max[i] else self.max[i]

            ptr = (
                (i * self._ctype_size + idx // self.type_size * self.byte_stride)
                + self.byte_offset
            )

            struct.pack_into(self._ctype, self._buffer_data, ptr, value)

    __slots__ = (
        "name",
        "bytelength",
        "buffer_views",
        "accessors",
        )

    def __init__(self, name):
        self.name = 'buffer_{}'.format(name)
        self.bytelength = 0
        self.buffer_views = collections.OrderedDict()
        self.accessors = {}

    def export_buffer(self, state):
        data = bytearray()
        for view in self.buffer_views.values():
            data.extend(view['data'])

        if state['settings']['buffers_embed_data'] and not state['settings']['gltf_export_binary']:
            uri = 'data:application/octet-stream;base64,' + base64.b64encode(data).decode('ascii')
        else:
            uri = bpy.path.clean_name(self.name) + '.bin'
            path = os.path.join(state['settings']['gltf_output_dir'], uri)
            state['files'][path] = data

        gltf = {
            'byteLength': self.bytelength,
            'name': self.name,
        }

        embedded_binary = (
            state['settings']['buffers_embed_data']
            and state['settings']['gltf_export_binary']
        )
        if not embedded_binary:
            gltf['uri'] = uri

        if state['version'] < Version('2.0'):
            gltf['type'] = 'arraybuffer'

        return gltf

    def add_view(self, bytelength, bytestride, target):
        buffer_name = 'bufferView_{}_{}'.format(self.name, len(self.buffer_views))
        self.buffer_views[buffer_name] = {
            'data': bytearray(bytelength),
            'target': target,
            'bytelength': bytelength,
            'byteoffset': self.bytelength,
            'bytestride': bytestride,
        }
        self.bytelength += bytelength
        return buffer_name

    def export_views(self, state):
        gltf_views = []

        for key, value in self.buffer_views.items():
            gltf = {
                'byteLength': value['bytelength'],
                'byteOffset': value['byteoffset'],
                'name': key,
            }

            if state['version'] >= Version('2.0') and value['bytestride'] > 0:
                gltf['byteStride'] = value['bytestride']

            gltf['buffer'] = Reference('buffers', self.name, gltf, 'buffer')
            state['references'].append(gltf['buffer'])

            if value['target'] is not None:
                gltf['target'] = value['target']

            gltf_views.append(gltf)

            state['input']['bufferViews'].append(SimpleID(key))

        return gltf_views

    def get_buffer_data(self, buffer_view):
        return self.buffer_views[buffer_view]['data']

    def add_accessor(self,
                     buffer_view,
                     byte_offset,
                     byte_stride,
                     component_type,
                     count,
                     data_type):
        accessor_name = 'accessor_{}_{}'.format(self.name, len(self.accessors))
        self.accessors[accessor_name] = self.Accessor(
            accessor_name,
            self,
            buffer_view,
            byte_offset,
            byte_stride,
            component_type,
            count,
            data_type
        )
        return self.accessors[accessor_name]

    def export_accessors(self, state):
        gltf_accessors = []

        for key, value in self.accessors.items():
            # Do not export an empty accessor
            if value.count == 0:
                continue

            gltf = {
                'byteOffset': value.byte_offset,
                'componentType': value.component_type,
                'count': value.count,
                'min': value.min[:value.type_size],
                'max': value.max[:value.type_size],
                'type': value.data_type,
                'name': value.name,
            }

            if state['version'] < Version('2.0'):
                gltf['byteStride'] = value.byte_stride

            gltf['bufferView'] = Reference('bufferViews', value.buffer_view, gltf, 'bufferView')
            state['references'].append(gltf['bufferView'])

            gltf_accessors.append(gltf)

            state['input']['accessors'].append(SimpleID(key))

        return gltf_accessors

    def combine(self, other, state):
        # Handle the simple stuff
        combined = Buffer(state['settings']['gltf_name'])
        combined.bytelength = self.bytelength + other.bytelength
        combined.accessors = {**self.accessors, **other.accessors}

        # Need to update byte offsets in buffer views
        combined.buffer_views = self.buffer_views.copy()
        other_views = other.buffer_views.copy()
        for key in other_views.keys():
            other_views[key]['byteoffset'] += self.bytelength
        combined.buffer_views.update(other_views)

        return combined



import argparse
import os
import sys
import tempfile


from .blend2gltf import ConverterBlend2Gltf
from .gltf2bam import ConverterGltf2Bam
from .common import Settings


def convert(settings, srcdir, src, dst):
    blend2gltf = ConverterBlend2Gltf(settings)
    gltf2bam = ConverterGltf2Bam(settings)

    for src_element in src:
        if not os.path.exists(src_element):
            print('Source ({}) does not exist'.format(src_element))
            sys.exit(1)

        if len(src) > 1 and not os.path.isfile(src_element):
            print('Source ({}) is not a file'.format(src_element))
            sys.exit(1)

        if len(src) == 1 and not (os.path.isfile(src_element) or os.path.isdir(src_element)):
            print('Source ({}) must be a file or a directory'.format(src))
            sys.exit(1)

    src_is_dir = os.path.isdir(src[0])
    dst_is_dir = not os.path.splitext(dst)[1]

    files_to_convert = []
    if src_is_dir:
        srcdir = src[0]
        for root, _, files in os.walk(srcdir):
            files_to_convert += [
                os.path.join(root, i)
                for i in files
                if i.endswith('.blend')
            ]
    else:
        files_to_convert = [os.path.abspath(i) for i in src]

    is_batch = len(files_to_convert) > 1

    if is_batch and not dst_is_dir:
        print('Destination must be a directory if the source is a directory or multiple files')

    if is_batch:
        # Batch conversion
        blend2gltf.convert_batch(srcdir, dst, files_to_convert)
        tmpfiles = [i.replace(srcdir, dst).replace('.blend', '.gltf') for i in files_to_convert]
        gltf2bam.convert_batch(dst, dst, tmpfiles)
        _ = [os.remove(i) for i in tmpfiles]
    else:
        # Single file conversion
        srcfile = files_to_convert[0]
        if dst_is_dir:
            # Destination is a directory, add a filename
            dst = os.path.join(dst, os.path.basename(srcfile.replace('blend', 'bam')))

        with tempfile.NamedTemporaryFile() as tmpfile:
            blend2gltf.convert_single(srcfile, tmpfile.name)
            gltf2bam.convert_single(tmpfile.name, dst)

def main():
    parser = argparse.ArgumentParser(
        description='CLI tool to convert Blender blend files to Panda3D BAM files'
    )

    parser.add_argument('src', nargs='+', type=str, help='source path')
    parser.add_argument('dst', type=str, help='destination path')

    parser.add_argument(
        '-m', '--material-mode',
        choices=[
            'legacy',
            'pbr',
        ],
        default='legacy',
        help='control how materials are exported'
    )

    parser.add_argument(
        '--srcdir',
        default=None,
        help='a common source directory to use when specifying multiple source files'
    )

    parser.add_argument(
        '--blender-dir',
        default='',
        help='directory that contains the blender binary'
    )

    args = parser.parse_args()

    src = [os.path.abspath(i) for i in args.src]
    srcdir = args.srcdir if args.srcdir is not None else os.path.commonpath(src)
    dst = os.path.abspath(args.dst)

    settings = Settings(
        material_mode=args.material_mode,
        blender_dir=args.blender_dir
    )

    convert(settings, srcdir, src, dst)



import importlib

from . import _lights_common

if '__IMPORTED__' in locals():
    importlib.reload(locals()['_lights_common'])
else:
    __IMPORTED__ = True


class KhrLights:
    ext_meta = {
        'name': 'KHR_lights',
        'url': (
            'https://github.com/andreasplesch/glTF/blob/ec6f61d73bcd58d59d4a4ea9ac009f973c693c5f/'
            'extensions/Khronos/KHR_lights/README.md'
        ),
        'isDraft': True,
    }

    def export(self, state):
        state['extensions_used'].append('KHR_lights')
        _lights_common.export_lights(state, 'KHR_lights')



from distutils.version import StrictVersion as Version
import functools
import itertools
import json
import struct

import bpy
import mathutils

from .exporters import (
    Buffer,
    Reference,
    SimpleID,
    get_bone_name,

    AnimationExporter,
    CameraExporter,
    ImageExporter,
    MaterialExporter,
    MeshExporter,
    NodeExporter,
    SceneExporter,
    SkinExporter,
    TextureExporter,
)


__all__ = ['export_gltf']


DEFAULT_SETTINGS = {
    'gltf_output_dir': '',
    'gltf_name': 'gltf',
    'gltf_export_binary': False,
    'buffers_embed_data': True,
    'buffers_combine_data': False,
    'nodes_export_hidden': False,
    'nodes_global_matrix': mathutils.Matrix.Identity(4),
    'nodes_global_matrix_apply': True,
    'nodes_selected_only': False,
    'blocks_prune_unused': True,
    'meshes_apply_modifiers': True,
    'meshes_interleave_vertex_data': True,
    'meshes_vertex_color_alpha': True,
    'images_data_storage': 'COPY',
    'asset_version': '2.0',
    'asset_copyright': '',
    'asset_profile': 'WEB',
    'images_allow_srgb': False,
    'extension_exporters': [],
    'animations_object_export': 'ACTIVE',
    'animations_armature_export': 'ELIGIBLE',
    'animations_shape_key_export': 'ELIGIBLE',
    'hacks_streaming': False,
}


PROFILE_MAP = {
    'WEB': {'api': 'WebGL', 'version': '1.0'},
    'DESKTOP': {'api': 'OpenGL', 'version': '3.0'}
}


def togl(matrix):
    return [i for col in matrix.col for i in col]


def _decompose(matrix):
    loc, rot, scale = matrix.decompose()
    loc = loc.to_tuple()
    rot = (rot.x, rot.y, rot.z, rot.w)
    scale = scale.to_tuple()

    return loc, rot, scale


def export_joint(state, bone):
    axis_mat = mathutils.Matrix.Identity(4)
    if state['settings']['nodes_global_matrix_apply']:
        axis_mat = state['settings']['nodes_global_matrix']

    matrix = axis_mat * bone.matrix_local
    if bone.parent:
        matrix = bone.parent.matrix_local.inverted() * bone.matrix_local

    bone_name = get_bone_name(bone)

    gltf_joint = {
        'name': bone.name,
    }
    if state['version'] < Version('2.0'):
        gltf_joint['jointName'] = Reference(
            'objects',
            bone_name,
            gltf_joint,
            'jointName'
        )
        state['references'].append(gltf_joint['jointName'])
    if bone.children:
        gltf_joint['children'] = [
            Reference('objects', get_bone_name(child), None, None) for child in bone.children
        ]
    if bone_name in state['bone_children']:
        bone_children = [
            Reference('objects', obj_name, None, None)
            for obj_name in state['bone_children'][bone_name]
        ]
        gltf_joint['children'] = gltf_joint.get('children', []) + bone_children
    for i, ref in enumerate(gltf_joint.get('children', [])):
        ref.source = gltf_joint['children']
        ref.prop = i
        state['references'].append(ref)

    (
        gltf_joint['translation'],
        gltf_joint['rotation'],
        gltf_joint['scale']
    ) = _decompose(matrix)

    return gltf_joint


def export_buffers(state):
    if state['settings']['buffers_combine_data']:
        buffers = [functools.reduce(
            lambda x, y: x.combine(y, state),
            state['buffers'],
            Buffer('empty')
        )]
        state['buffers'] = buffers
        state['input']['buffers'] = [SimpleID(buffers[0].name)]
    else:
        buffers = state['buffers']

    gltf = {}
    gltf['buffers'] = [buf.export_buffer(state) for buf in buffers]
    gltf['bufferViews'] = list(itertools.chain(*[buf.export_views(state) for buf in buffers]))
    gltf['accessors'] = list(itertools.chain(*[buf.export_accessors(state) for buf in buffers]))

    return gltf


def insert_root_nodes(state, root_matrix):
    for i, scene in enumerate(state['output']['scenes']):
        # Generate a new root node for each scene
        root_node = {
            'children': scene['nodes'],
            'matrix': root_matrix,
            'name': '{}_root'.format(scene['name']),
        }
        state['output']['nodes'].append(root_node)
        ref_name = '__scene_root_{}_'.format(i)
        state['input']['objects'].append(SimpleID(ref_name))

        # Replace scene node lists to just point to the new root nodes
        scene['nodes'] = []
        scene['nodes'].append(Reference('objects', ref_name, scene['nodes'], 0))
        state['references'].append(scene['nodes'][0])


def build_string_refmap(input_data):
    in_out_map = {
        'objects': 'nodes',
        'bones': 'nodes',
        'lamps': 'lights'
    }
    refmap = {}
    for key, value in input_data.items():
        refmap.update({
            (key, data.name): '{}_{}'.format(in_out_map.get(key, key), data.name)
            for data in value
        })
    return refmap


def build_int_refmap(input_data):
    refmap = {}
    for key, value in input_data.items():
        refmap.update({(key, data.name): i for i, data in enumerate(value)})
    return refmap


def initialize_state(settings=None):
    # Fill in any missing settings with defaults
    if not settings:
        settings = {}
    for key, value in DEFAULT_SETTINGS.items():
        settings.setdefault(key, value)

    res_x = bpy.context.scene.render.resolution_x
    res_y = bpy.context.scene.render.resolution_y
    # Initialize export state
    state = {
        'version': Version(settings['asset_version']),
        'settings': settings,
        'animation_dt': 1.0 / bpy.context.scene.render.fps,
        'mod_meshes_obj': {},
        'aspect_ratio': res_x / res_y,
        'mod_meshes': {},
        'shape_keys': {},
        'dupli_nodes': [],
        'bone_children': {},
        'skinned_meshes': set(),
        'extensions_used': [],
        'gl_extensions_used': [],
        'buffers': [],
        'samplers': [],
        'input': {
            'buffers': [],
            'accessors': [],
            'bufferViews': [],
            'objects': [],
            'bones': [],
            'anim_samplers': [],
            'samplers': [],
            'scenes': [],
            'skins': [],
            'materials': [],
            'dupli_ids': [],
        },
        'output': {
            'extensions': [],
        },
        'references': [],
        'files': {},
        'decompose_fn': _decompose,
        'decompose_mesh_fn': _decompose,
    }

    return state


def export_gltf(scene_delta, settings=None):
    state = initialize_state(settings)
    state['input'].update({key: list(value) for key, value in scene_delta.items()})

    # Filter out empty meshes
    if 'meshes' in state['input']:
        state['input']['meshes'] = [mesh for mesh in state['input']['meshes'] if mesh.loops]
        if 'objects' in state['input']:
            state['input']['objects'] = [
                obj for obj in state['input']['objects']
                if obj.type != 'MESH' or obj.data in state['input']['meshes']
            ]

    # Make sure any temporary meshes do not have animation data baked in
    default_scene = bpy.context.scene
    if not state['settings']['hacks_streaming']:
        saved_pose_positions = [armature.pose_position for armature in bpy.data.armatures]
        for armature in bpy.data.armatures:
            armature.pose_position = 'REST'
        if saved_pose_positions:
            for obj in bpy.data.objects:
                if obj.type == 'ARMATURE':
                    obj.update_tag()
        default_scene.frame_set(default_scene.frame_current)

    mesh_list = []
    mod_obs = [
        ob for ob in state['input']['objects']
        if [mod for mod in ob.modifiers if mod.type != 'ARMATURE']
    ]
    for mesh in state['input'].get('meshes', []):
        if mesh.shape_keys and mesh.shape_keys.use_relative:
            relative_key = mesh.shape_keys.key_blocks[0].relative_key
            keys = [key for key in mesh.shape_keys.key_blocks if key != relative_key]

            # Gather weight values
            weights = [key.value for key in keys]

            # Clear weight values
            for key in keys:
                key.value = 0.0
            mesh_users = [obj for obj in state['input']['objects'] if obj.data == mesh]

            # Mute modifiers if necessary
            muted_modifiers = []
            original_modifier_states = []
            if not state['settings']['meshes_apply_modifiers']:
                muted_modifiers = itertools.chain.from_iterable(
                    [obj.modifiers for obj in mesh_users]
                )
                original_modifier_states = [mod.show_viewport for mod in muted_modifiers]
                for modifier in muted_modifiers:
                    modifier.show_viewport = False

            for user in mesh_users:
                base_mesh = user.to_mesh(default_scene, True, 'PREVIEW')
                mesh_name = base_mesh.name
                state['mod_meshes_obj'][user.name] = base_mesh

                if mesh_name not in state['shape_keys']:
                    key_meshes = []
                    for key, weight in zip(keys, weights):
                        key.value = key.slider_max
                        key_meshes.append((
                            weight,
                            user.to_mesh(default_scene, True, 'PREVIEW')
                        ))
                        key.value = 0.0
                    state['shape_keys'][mesh_name] = key_meshes

            # Reset weight values
            for key, weight in zip(keys, weights):
                key.value = weight

            # Unmute modifiers
            for modifier, state in zip(muted_modifiers, original_modifier_states):
                modifier.show_viewport = state
        elif state['settings']['meshes_apply_modifiers']:
            mod_users = [ob for ob in mod_obs if ob.data == mesh]

            # Only convert meshes with modifiers, otherwise each non-modifier
            # user ends up with a copy of the mesh and we lose instancing
            state['mod_meshes_obj'].update(
                {ob.name: ob.to_mesh(default_scene, True, 'PREVIEW') for ob in mod_users}
            )

            # Add unmodified meshes directly to the mesh list
            if len(mod_users) < mesh.users:
                mesh_list.append(mesh)
        else:
            mesh_list.append(mesh)

    mesh_list.extend(state['mod_meshes_obj'].values())
    state['input']['meshes'] = mesh_list

    apply_global_matrix = (
        state['settings']['nodes_global_matrix'] != mathutils.Matrix.Identity(4)
        and state['settings']['nodes_global_matrix_apply']
    )
    if apply_global_matrix:
        global_mat = state['settings']['nodes_global_matrix']
        global_scale_mat = mathutils.Matrix([[abs(j) for j in i] for i in global_mat])

        def decompose_apply(matrix):
            loc, rot, scale = matrix.decompose()

            loc.rotate(global_mat)
            loc = loc.to_tuple()

            rot.rotate(global_mat)
            rot = (rot.x, rot.y, rot.z, rot.w)

            scale.rotate(global_scale_mat)
            scale = scale.to_tuple()

            return loc, rot, scale

        def decompose_mesh_apply(matrix):
            loc, rot, scale = matrix.decompose()

            loc.rotate(global_mat)
            loc = loc.to_tuple()

            rot = mathutils.Vector(list(rot.to_euler()))
            rot.rotate(global_mat)
            rot = mathutils.Euler(rot, 'XYZ').to_quaternion()
            rot = (rot.x, rot.y, rot.z, rot.w)

            scale.rotate(global_scale_mat)
            scale = scale.to_tuple()

            return loc, rot, scale
        state['decompose_fn'] = decompose_apply
        state['decompose_mesh_fn'] = decompose_mesh_apply

        transformed_meshes = [mesh.copy() for mesh in mesh_list]
        for mesh in transformed_meshes:
            mesh.transform(global_mat, shape_keys=False)
        state['mod_meshes'].update(
            {mesh.name: xformed_mesh for xformed_mesh, mesh in zip(transformed_meshes, mesh_list)}
        )
        for shape_key_list in state['shape_keys'].values():
            for shape_key in shape_key_list:
                shape_key[1].transform(global_mat, shape_keys=False)

    # Restore armature pose positions
    for i, armature in enumerate(bpy.data.armatures):
        armature.pose_position = saved_pose_positions[i]

    state['input']['action_pairs'] = AnimationExporter.gather(state)

    exporters = [
        CameraExporter,
        ImageExporter,
        NodeExporter,
        MaterialExporter,
        # Make sure meshes come after nodes to detect which meshes are skinned
        MeshExporter,
        SceneExporter,
        SkinExporter,
        TextureExporter,
        AnimationExporter,
    ]
    state['output'] = {
        exporter.gltf_key: [
            exporter.export(state, data)
            if exporter.check(state, data)
            else exporter.default(state, data)
            for data in state['input'].get(exporter.blender_key, [])
        ] for exporter in exporters
    }

    # Export top level data
    gltf = {
        'asset': {
            'version': state['settings']['asset_version'],
            'generator': 'blendergltf v1.2.0',
            'copyright': state['settings']['asset_copyright'],
        }
    }
    if state['version'] < Version('2.0'):
        gltf['asset']['profile'] = PROFILE_MAP[state['settings']['asset_profile']]

    # Export samplers
    state['output']['samplers'] = state['samplers']

    # Export animations
    state['output']['nodes'].extend([
        export_joint(state, sid.data) for sid in state['input']['bones']
    ])

    # Move bones to nodes for updating references
    state['input']['objects'].extend(state['input']['bones'])
    state['input']['bones'] = []

    # Export dupli-groups
    state['output']['nodes'].extend(state['dupli_nodes'])
    state['input']['objects'].extend(state['input']['dupli_ids'])
    state['input']['dupli_ids'] = []

    # Export default scene
    default_scene = None
    for scene in state['input']['scenes']:
        if scene == bpy.context.scene:
            default_scene = scene
    if default_scene:
        scene_ref = Reference('scenes', bpy.context.scene.name, gltf, 'scene')
        scene_ref.value = 0
        state['references'].append(scene_ref)

    # Export extensions
    state['refmap'] = build_int_refmap(state['input'])
    for ext_exporter in state['settings']['extension_exporters']:
        ext_exporter.export(state)

    # Insert root nodes if axis conversion is needed
    root_node_needed = (
        state['settings']['nodes_global_matrix'] != mathutils.Matrix.Identity(4)
        and not state['settings']['nodes_global_matrix_apply']
    )
    if root_node_needed:
        insert_root_nodes(state, togl(state['settings']['nodes_global_matrix']))

    if state['buffers']:
        state['output'].update(export_buffers(state))
    state['output'] = {key: value for key, value in state['output'].items() if value != []}
    if state['extensions_used']:
        gltf.update({'extensionsUsed': state['extensions_used']})
    if state['version'] < Version('2.0'):
        gltf.update({'glExtensionsUsed': state['gl_extensions_used']})

    # Convert lists to dictionaries
    if state['version'] < Version('2.0'):
        extensions = state['output'].get('extensions', [])
        state['output'] = {
            key: {
                '{}_{}'.format(key, data['name']): data for data in value
            } for key, value in state['output'].items()
            if key != 'extensions'
        }
        if extensions:
            state['output']['extensions'] = extensions
    gltf.update(state['output'])

    # Gather refmap inputs
    reference_inputs = state['input']
    if state['settings']['hacks_streaming']:
        reference_inputs.update({
            'actions': list(bpy.data.actions),
            'cameras': list(bpy.data.cameras),
            'lamps': list(bpy.data.lamps),
            'images': list(bpy.data.images),
            'materials': list(bpy.data.materials),
            'meshes': list(bpy.data.meshes),
            'objects': list(bpy.data.objects),
            'scenes': list(bpy.data.scenes),
            'textures': list(bpy.data.textures),
        })

    # Resolve references
    if state['version'] < Version('2.0'):
        refmap = build_string_refmap(reference_inputs)
        ref_default = 'INVALID'
    else:
        refmap = build_int_refmap(reference_inputs)
        ref_default = -1
    for ref in state['references']:
        ref.source[ref.prop] = refmap.get((ref.blender_type, ref.blender_name), ref_default)
        if ref.source[ref.prop] == ref_default:
            print(
                'Warning: {} contains an invalid reference to {}'
                .format(ref.source, (ref.blender_type, ref.blender_name))
            )

    # Remove any temporary meshes
    temp_mesh_collections = (
        state['mod_meshes'].values(),
        state['mod_meshes_obj'].values(),
        [
            shape_key_pair[1] for shape_key_pair
            in itertools.chain.from_iterable(state['shape_keys'].values())
        ]
    )
    for mesh in itertools.chain(*temp_mesh_collections):
        bpy.data.meshes.remove(mesh)

    # Transform gltf data to binary
    if state['settings']['gltf_export_binary']:
        json_data = json.dumps(gltf, sort_keys=True, check_circular=False).encode()
        json_length = len(json_data)
        json_pad = (' ' * (4 - json_length % 4)).encode()
        json_pad = json_pad if len(json_pad) != 4 else b''
        json_length += len(json_pad)
        json_format = '<II{}s{}s'.format(len(json_data), len(json_pad))
        chunks = [struct.pack(json_format, json_length, 0x4e4f534a, json_data, json_pad)]

        if settings['buffers_embed_data']:
            buffers = [data for path, data in state['files'].items() if path.endswith('.bin')]

            # Get padded lengths
            lengths = [len(buffer) for buffer in buffers]
            lengths = [
                length + ((4 - length % 4) if length % 4 != 0 else 0)
                for length in lengths
            ]

            chunks.extend([
                struct.pack('<II{}s'.format(length), length, 0x004E4942, buffer)
                for buffer, length in zip(buffers, lengths)
            ])

            state['files'] = {
                path: data for path, data in state['files'].items()
                if not path.endswith('.bin')
            }

        version = 2
        size = 12
        for chunk in chunks:
            size += len(chunk)
        header = struct.pack('<4sII', b'glTF', version, size)

        gltf = bytes(0).join([header, *chunks])

    # Write secondary files
    for path, data in state['files'].items():
        with open(path, 'wb') as fout:
            fout.write(data)

    return gltf



import importlib

from ..blendergltf import Reference

from . import _lights_common

if '__IMPORTED__' in locals():
    importlib.reload(locals()['_lights_common'])
else:
    __IMPORTED__ = True


class KhrMaterialsCommon:
    ext_meta = {
        'name': 'KHR_materials_common',
        'url': (
            'https://github.com/KhronosGroup/glTF/tree/master/extensions/'
            'Khronos/KHR_materials_common'
        ),
        'isDraft': True,
    }

    def export_material(self, state, material):
        all_textures = [
            slot for slot in material.texture_slots
            if slot and slot.texture.type == 'IMAGE'
        ]
        diffuse_textures = [
            Reference('textures', t.texture.name, None, None)
            for t in all_textures if t.use_map_color_diffuse
        ]
        emission_textures = [
            Reference('textures', t.texture.name, None, None)
            for t in all_textures
            if (
                (material.use_shadeless and t.use_map_color_diffuse)
                or (not material.use_shadeless and t.use_map_emit)
            )
        ]
        specular_textures = [
            Reference('textures', t.texture.name, None, None)
            for t in all_textures if t.use_map_color_spec
        ]

        diffuse_color = list((material.diffuse_color * material.diffuse_intensity)[:])
        diffuse_color += [material.alpha]
        emission_color = list((material.diffuse_color * material.emit)[:])
        emission_color += [material.alpha]
        specular_color = list((material.specular_color * material.specular_intensity)[:])
        specular_color += [material.specular_alpha]

        technique = 'PHONG'
        if material.use_shadeless:
            technique = 'CONSTANT'
            emission_color = diffuse_color
        elif material.specular_intensity == 0.0:
            technique = 'LAMBERT'
        elif material.specular_shader == 'BLINN':
            technique = 'BLINN'

        gltf = {
            'technique': technique,
            'values': {
                'ambient': ([material.ambient]*3) + [1.0],
                'diffuse': diffuse_textures[-1] if diffuse_textures else diffuse_color,
                'doubleSided': not material.game_settings.use_backface_culling,
                'emission': emission_textures[-1] if emission_textures else emission_color,
                'specular': specular_textures[-1] if specular_textures else specular_color,
                'shininess': material.specular_hardness,
                'transparency': material.alpha,
                'transparent': material.use_transparency,
            }
        }

        for prop in ('diffuse', 'emission', 'specular'):
            if hasattr(gltf['values'][prop], 'blender_type'):
                ref = gltf['values'][prop]
                ref.source = gltf['values']
                ref.prop = prop
                state['references'].append(ref)

        return gltf

    def export(self, state):
        state['extensions_used'].append('KHR_materials_common')

        # Export materials
        material_pairs = [
            (material, state['output']['materials'][state['refmap'][('materials', material.name)]])
            for material in state['input']['materials']
        ]
        for bl_mat, gl_mat in material_pairs:
            gl_mat['extensions'] = gl_mat.get('extensions', {})
            gl_mat['extensions']['KHR_materials_common'] = self.export_material(state, bl_mat)

        _lights_common.export_lights(state, 'KHR_materials_common')



from distutils.version import StrictVersion as Version
import itertools

import bpy
import mathutils

from .base import BaseExporter
from .common import (
    AnimationPair,
    Buffer,
    Reference,
    SimpleID,
    get_bone_name,
)


def _decompose(matrix):
    loc, rot, scale = matrix.decompose()
    loc = loc.to_tuple()
    rot = (rot.x, rot.y, rot.z, rot.w)
    scale = scale.to_tuple()

    return loc, rot, scale


def _can_object_use_action(obj, action):
    for fcurve in action.fcurves:
        path = fcurve.data_path
        if not path.startswith('pose'):
            return obj.animation_data is not None

        if obj.type == 'ARMATURE':
            path = path.split('["')[-1]
            path = path.split('"]')[0]
            if path in [bone.name for bone in obj.data.bones]:
                return True

    return False


class AnimationExporter(BaseExporter):
    gltf_key = 'animations'
    blender_key = 'action_pairs'

    @classmethod
    def gather_actions(cls, state):
        armature_objects = [obj for obj in state['input']['objects'] if obj.type == 'ARMATURE']
        regular_objects = [obj for obj in state['input']['objects'] if obj.type != 'ARMATURE']

        actions = []

        def export_eligible(objects):
            for obj in objects:
                actions.extend([
                    AnimationPair(obj, action) for action in state['input']['actions']
                    if _can_object_use_action(obj, action)
                ])

        def export_active(objects):
            for obj in objects:
                if obj.animation_data and obj.animation_data.action:
                    actions.append(AnimationPair(obj, obj.animation_data.action))

        armature_setting = state['settings']['animations_armature_export']
        object_setting = state['settings']['animations_object_export']

        if armature_setting == 'ACTIVE':
            export_active(armature_objects)
        elif armature_setting == 'ELIGIBLE':
            export_eligible(armature_objects)
        else:
            print(
                'WARNING: Unrecognized setting for animations_armature_export:',
                '{}'.format(armature_setting)
            )

        if object_setting == 'ACTIVE':
            export_active(regular_objects)
        elif object_setting == 'ELIGIBLE':
            export_eligible(regular_objects)
        else:
            print(
                'WARNING: Unrecognized setting for animations_object_export:',
                '{}'.format(object_setting)
            )

        return actions

    @classmethod
    def gather_shape_key_actions(cls, state):
        shape_key_objects = [
            obj for obj in state['input']['objects']
            if obj.type == 'MESH' and obj.data.shape_keys
        ]
        shape_key_setting = state['settings']['animations_shape_key_export']

        shape_key_actions = []

        if shape_key_setting == 'ACTIVE':
            for obj in shape_key_objects:
                action = obj.data.shape_keys.animation_data.action
                shape_key_actions.append(AnimationPair(obj, action, True))
        elif shape_key_setting == 'ELIGIBLE':
            for obj in shape_key_objects:
                eligible_actions = []
                shape_keys = set([
                    block.name for block in obj.data.shape_keys.key_blocks
                    if block != block.relative_key
                ])
                for action in state['input']['actions']:
                    fcurve_keys = set([fcurve.data_path.split('"')[1] for fcurve in action.fcurves])
                    if fcurve_keys <= shape_keys:
                        eligible_actions.append(action)
                for action in eligible_actions:
                    shape_key_actions.append(AnimationPair(obj, action, True))
        else:
            print(
                'WARNING: Unrecognized setting for animations_shape_key_export:',
                '{}'.format(shape_key_setting)
            )

        return shape_key_actions

    @classmethod
    def gather(cls, state):
        actions = cls.gather_actions(state)
        actions.extend(cls.gather_shape_key_actions(state))
        return actions

    @classmethod
    def export_action(cls, state, blender_data):
        obj = blender_data.target
        action = blender_data.action
        action_name = blender_data.name

        if state['version'] < Version('2.0'):
            target_key = 'id'
        else:
            target_key = 'node'

        channels = {}
        decompose = state['decompose_fn']
        axis_mat = mathutils.Matrix.Identity(4)
        if state['settings']['nodes_global_matrix_apply']:
            axis_mat = state['settings']['nodes_global_matrix']

        sce = bpy.context.scene
        prev_frame = sce.frame_current
        prev_action = obj.animation_data.action

        frame_start, frame_end = [int(x) for x in action.frame_range]
        num_frames = frame_end - frame_start + 1
        obj.animation_data.action = action

        has_location = set()
        has_rotation = set()
        has_scale = set()

        # Check action groups to see what needs to be animated
        pose_bones = {}
        for group in action.groups:
            for channel in group.channels:
                data_path = channel.data_path
                if obj.pose and 'pose.bones' in data_path:
                    target_name = data_path.split('"')[1]
                    transform = data_path.split('.')[-1]
                    pose_bones[obj.pose.bones[target_name]] = None
                else:
                    target_name = obj.name
                    transform = data_path.lower()
                    if obj.name not in channels:
                        channels[obj.name] = []

                if 'location' in transform:
                    has_location.add(target_name)
                if 'rotation' in transform:
                    has_rotation.add(target_name)
                if 'scale' in transform:
                    has_scale.add(target_name)
        channels.update({pbone.name: [] for pbone in pose_bones})

        # Iterate frames and bake animations
        for frame in range(frame_start, frame_end + 1):
            sce.frame_set(frame)

            if obj.name in channels:
                # Decompose here so we don't store a reference to the matrix
                loc, rot, scale = decompose(obj.matrix_local)
                if obj.name not in has_location:
                    loc = None
                if obj.name not in has_rotation:
                    rot = None
                if obj.name not in has_scale:
                    scale = None
                channels[obj.name].append((loc, rot, scale))

            for pbone in pose_bones:
                if pbone.parent:
                    mat = pbone.parent.matrix.inverted() * pbone.matrix
                else:
                    mat = axis_mat * pbone.matrix

                loc, rot, scale = _decompose(mat)

                if pbone.name not in has_location:
                    loc = None
                if pbone.name not in has_rotation:
                    rot = None
                if pbone.name not in has_scale:
                    scale = None
                channels[pbone.name].append((loc, rot, scale))

        gltf_channels = []
        gltf_parameters = {}
        gltf_samplers = []

        tbuf = Buffer('{}_time'.format(action_name))
        tbv = tbuf.add_view(num_frames * 1 * 4, 1 * 4, None)
        tdata = tbuf.add_accessor(tbv, 0, 1 * 4, Buffer.FLOAT, num_frames, Buffer.SCALAR)
        time = 0
        for i in range(num_frames):
            tdata[i] = time
            time += state['animation_dt']
        state['buffers'].append(tbuf)
        state['input']['buffers'].append(SimpleID(tbuf.name))
        time_parameter_name = '{}_time_parameter'.format(action_name)
        ref = Reference('accessors', tdata.name, gltf_parameters, time_parameter_name)
        gltf_parameters[time_parameter_name] = ref
        state['references'].append(ref)

        input_list = '{}_{}_samplers'.format(action_name, obj.name)
        state['input'][input_list] = []

        sampler_keys = []
        for targetid, chan in channels.items():
            buf = Buffer('{}_{}'.format(targetid, action_name))
            ldata = rdata = sdata = None
            paths = []
            if targetid in has_location:
                lbv = buf.add_view(num_frames * 3 * 4, 3 * 4, None)
                ldata = buf.add_accessor(lbv, 0, 3 * 4, Buffer.FLOAT, num_frames, Buffer.VEC3)
                paths.append('translation')
            if targetid in has_rotation:
                rbv = buf.add_view(num_frames * 4 * 4, 4 * 4, None)
                rdata = buf.add_accessor(rbv, 0, 4 * 4, Buffer.FLOAT, num_frames, Buffer.VEC4)
                paths.append('rotation')
            if targetid in has_scale:
                sbv = buf.add_view(num_frames * 3 * 4, 3 * 4, None)
                sdata = buf.add_accessor(sbv, 0, 3 * 4, Buffer.FLOAT, num_frames, Buffer.VEC3)
                paths.append('scale')

            if not paths:
                continue

            for i in range(num_frames):
                loc, rot, scale = chan[i]
                if ldata:
                    for j in range(3):
                        ldata[(i * 3) + j] = loc[j]
                if sdata:
                    for j in range(3):
                        sdata[(i * 3) + j] = scale[j]
                if rdata:
                    for j in range(4):
                        rdata[(i * 4) + j] = rot[j]

            state['buffers'].append(buf)
            state['input']['buffers'].append(SimpleID(buf.name))

            is_bone = False
            if targetid != obj.name:
                is_bone = True
                targetid = get_bone_name(bpy.data.armatures[obj.data.name].bones[targetid])

            for path in paths:
                sampler_name = '{}_{}_{}_sampler'.format(action_name, targetid, path)
                sampler_keys.append(sampler_name)
                parameter_name = '{}_{}_{}_parameter'.format(action_name, targetid, path)

                gltf_channel = {
                    'sampler': sampler_name,
                    'target': {
                        target_key: targetid,
                        'path': path,
                    }
                }
                gltf_channels.append(gltf_channel)
                id_ref = Reference(
                    'objects' if is_bone else 'objects',
                    targetid,
                    gltf_channel['target'],
                    target_key
                )
                state['references'].append(id_ref)
                state['input'][input_list].append(SimpleID(sampler_name))
                sampler_ref = Reference(input_list, sampler_name, gltf_channel, 'sampler')
                state['references'].append(sampler_ref)

                gltf_sampler = {
                    'input': None,
                    'interpolation': 'LINEAR',
                    'output': None,
                }
                gltf_samplers.append(gltf_sampler)

                accessor_name = {
                    'translation': ldata.name if ldata else None,
                    'rotation': rdata.name if rdata else None,
                    'scale': sdata.name if sdata else None,
                }[path]

                if state['version'] < Version('2.0'):
                    gltf_sampler['input'] = time_parameter_name
                    gltf_sampler['output'] = parameter_name
                    accessor_ref = Reference(
                        'accessors',
                        accessor_name,
                        gltf_parameters,
                        parameter_name
                    )
                    gltf_parameters[parameter_name] = accessor_ref
                else:
                    time_ref = Reference(
                        'accessors',
                        tdata.name,
                        gltf_sampler,
                        'input'
                    )
                    gltf_sampler['input'] = time_ref
                    state['references'].append(time_ref)
                    accessor_ref = Reference(
                        'accessors',
                        accessor_name,
                        gltf_sampler,
                        'output'
                    )
                    gltf_sampler['output'] = accessor_ref

                state['references'].append(accessor_ref)

        gltf_action = {
            'name': blender_data.action.name,
            'channels': gltf_channels,
            'samplers': gltf_samplers,
        }

        if state['version'] < Version('2.0'):
            gltf_action['samplers'] = {
                '{}_{}'.format(input_list, i[0]): i[1]
                for i in zip(sampler_keys, gltf_action['samplers'])
            }
            gltf_action['parameters'] = gltf_parameters

        sce.frame_set(prev_frame)
        obj.animation_data.action = prev_action

        return gltf_action

    @classmethod
    def export_shape_key_action(cls, state, blender_data):
        obj = blender_data.target
        action = blender_data.action
        action_name = blender_data.name

        fcurves = action.fcurves
        frame_range = action.frame_range
        frame_count = int(frame_range[1]) - int(frame_range[0])
        for fcurve in fcurves:
            fcurve.convert_to_samples(*frame_range)
        samples = {
            fcurve.data_path.split('"')[1]: [point.co[1] for point in fcurve.sampled_points]
            for fcurve in fcurves
        }
        shape_keys = [
            block for block in obj.data.shape_keys.key_blocks
            if block != block.relative_key
        ]
        empty_data = [0.0] * frame_count

        weight_data = zip(*[samples.get(key.name, empty_data) for key in shape_keys])
        weight_data = itertools.chain.from_iterable(weight_data)
        dt_data = [state['animation_dt'] * i for i in range(frame_count)]

        anim_buffer = Buffer('{}_{}'.format(obj.name, action_name))
        state['buffers'].append(anim_buffer)
        state['input']['buffers'].append(SimpleID(anim_buffer.name))

        time_view = anim_buffer.add_view(frame_count * 1 * 4, 1 * 4, None)
        time_acc = anim_buffer.add_accessor(
            time_view,
            0,
            1 * 4,
            Buffer.FLOAT,
            frame_count,
            Buffer.SCALAR
        )
        for i, dt in enumerate(dt_data):
            time_acc[i] = dt

        key_count = len(shape_keys)
        weight_view = anim_buffer.add_view(frame_count * key_count * 4, 4, None)
        weight_acc = anim_buffer.add_accessor(
            weight_view,
            0,
            1 * 4,
            Buffer.FLOAT,
            frame_count * key_count,
            Buffer.SCALAR
        )
        for i, weight in enumerate(weight_data):
            weight_acc[i] = weight

        channel = {
            'sampler': 0,
            'target': {
                'path': 'weights',
            },
        }
        channel['target']['node'] = Reference('objects', obj.name, channel['target'], 'node')
        state['references'].append(channel['target']['node'])

        sampler = {
            'interpolation': 'LINEAR',
        }
        sampler['input'] = Reference('accessors', time_acc.name, sampler, 'input')
        state['references'].append(sampler['input'])
        sampler['output'] = Reference('accessors', weight_acc.name, sampler, 'output')
        state['references'].append(sampler['output'])

        gltf_action = {
            'name': blender_data.action.name,
            'channels': [channel],
            'samplers': [sampler],
        }

        return gltf_action

    @classmethod
    def export(cls, state, blender_data):
        if blender_data.is_shape_key:
            return cls.export_shape_key_action(state, blender_data)

        return cls.export_action(state, blender_data)



import mathutils


class AttributeData:
    def __init__(self, mesh):
        self.mesh = mesh
        self.restart()

    def restart(self):
        self.positions = PositionsData(self.mesh)
        self.normals = NormalsData(self.mesh)
        self.group = GroupsData(self.mesh)
        self.color_layers = [ColorsData(c) for c in self.mesh.vertex_colors]
        self.uv_layers = [UvsData(l) for l in self.mesh.uv_layers]

        materials = [None] + list(self.mesh.materials)
        self.triangle_sets = [
            IndicesData(self.mesh, i, triangulate=True) for i, _ in enumerate(materials)
        ]
        self._prune_no_material_sets(self.triangle_sets)

        self.polygon_sets = [IndicesData(self.mesh, i) for i, _ in enumerate(materials)]
        self._prune_no_material_sets(self.polygon_sets)

    def _prune_no_material_sets(self, primitive_set):
        # pylint: disable=len-as-condition
        if not len(primitive_set[-1]):
            primitive_set.pop()


class CollectionData:
    def __init__(self, source, map_function):
        self.source = source
        self.iterator = (map_function(i) for i in source)

    def __len__(self):
        return len(self.source)


class PositionsData(CollectionData):
    def __init__(self, mesh):
        super().__init__(mesh.vertices, lambda v: v.co)


class GroupsData(CollectionData):
    def __init__(self, mesh):
        super().__init__(mesh.vertices, lambda v: v.groups)


class NormalsData(CollectionData):
    def __init__(self, mesh):
        mesh.calc_normals()
        super().__init__(mesh.vertices, lambda v: v.normal)


class ColorsData(CollectionData):
    # TODO: Address color space
    def __init__(self, colors):
        super().__init__(colors.data, lambda data: data.color)
        self.layer = colors


class UvsData(CollectionData):
    def __init__(self, uv_layer):
        super().__init__(uv_layer.data, lambda data: data.uv)
        self.layer = uv_layer


class IndicesData:
    def __init__(self, mesh, material_index, triangulate=False):
        self.mesh = mesh
        self.material = None
        if material_index < len(mesh.materials):
            self.material = mesh.materials[material_index]
        self._faces = [f for f in mesh.polygons if f.material_index == material_index]
        if triangulate:
            self._len = sum([max(len(f.vertices) - 2, 1) for f in self._faces])
            self.iterator = self._create_iter()
        else:
            self._len = len(self._faces)
            self.iterator = (p for p in self._faces)

    def __len__(self):
        return self._len

    def _create_iter(self):
        for face in self._faces:
            vertices = face.vertices
            if len(vertices) < 3:
                continue
            elif len(vertices) > 3:
                coords = [self.mesh.vertices[i].co for i in vertices]
                triangles = mathutils.geometry.tessellate_polygon((coords,))
                for triangle in triangles:
                    yield [vertices[i] for i in triangle[:1] + triangle[:0:-1]]
            else:
                yield list(vertices)


def extract_attributes(mesh):
    return AttributeData(mesh)



import json


def _is_serializable(value):
    try:
        json.dumps(value)
        return True
    except TypeError:
        return False


_IGNORED_CUSTOM_PROPS = [
    '_RNA_UI',
    'cycles',
    'cycles_visibility',
]


# pylint: disable=unused-argument
class BaseExporter:
    gltf_key = ''
    blender_key = ''

    @classmethod
    def get_custom_properties(cls, blender_data):
        custom_props = {
            key: value.to_list() if hasattr(value, 'to_list') else value
            for key, value in blender_data.items()
            if key not in _IGNORED_CUSTOM_PROPS
        }

        custom_props = {
            key: value for key, value in custom_props.items()
            if _is_serializable(value)
        }

        return custom_props

    @classmethod
    def check(cls, state, blender_data):
        return True

    @classmethod
    def default(cls, state, blender_data):
        return {
            'name': blender_data.name
        }

    @classmethod
    def export(cls, state, blender_data):
        return {}



import math

import bpy
import mathutils

ALPHA_MODE_ITEMS = [
    ('OPAQUE', 'Opaque', 'The alpha value is ignored and the rendered output is fully opaque'),
    ('MASK', 'Mask', (
        'The rendered output is either fully opaque or fully transparent depending on the '
        'alpha value and the specified alpha cutoff value'
    )),
    ('BLEND', 'Blend', 'The alpha value is used to composite the source and destination areas'),
]


def get_base_color_factor(self):
    material = self.id_data
    diffuse = mathutils.Vector(material.diffuse_color)
    diffuse *= material.diffuse_intensity
    return [*diffuse, material.alpha]


def set_base_color_factor(self, value):
    material = self.id_data
    material.diffuse_color = value[:3]
    material.diffuse_intensity = 1.0

    alpha = value[3]
    material.alpha = alpha
    if alpha < 1.0:
        material.use_transparency = True
        material.transparency_method = 'Z_TRANSPARENCY'
    else:
        material.use_transparency = False


def get_emissive_factor(self):
    material = self.id_data
    return [min(material.emit, 2.0) * 0.5] * 3


def set_emissive_factor(self, value):
    material = self.id_data
    material.emit = mathutils.Color(value).v * 2.0


def get_alpha_mode(self):
    material = self.id_data

    alpha_mode = 'OPAQUE'

    if material.use_transparency:
        gs_alpha = material.game_settings.alpha_blend
        if gs_alpha == 'CLIP':
            alpha_mode = 'MASK'
        else:
            alpha_mode = 'BLEND'

    for i, mode in enumerate(ALPHA_MODE_ITEMS):
        if mode[0] == alpha_mode:
            alpha_mode = i
            break
    else:
        alpha_mode = 0

    return alpha_mode


def set_alpha_mode(self, value):
    material = self.id_data

    value = ALPHA_MODE_ITEMS[value][0]

    if value == 'OPAQUE':
        material.use_transparency = False
        material.game_settings.alpha_blend = 'OPAQUE'
    elif value == 'MASK':
        material.use_transparency = True
        material.game_settings.alpha_blend = 'CLIP'
    elif value == 'BLEND':
        material.use_transparency = True
        material.game_settings.alpha_blend = 'ALPHA'


def get_roughness_factor(self):
    material = self.id_data
    hardness = material.specular_hardness
    if 1.0 < self.hardness_float < 511.0 and not hardness < self.hardness_float < hardness + 1:
        self.hardness_float = material.specular_hardness
    roughness = pow(2.0 / (self.hardness_float + 2.0), 0.25)
    return max(min(roughness, 1.0), 0.0)


def set_roughness_factor(self, value):
    material = self.id_data
    if value <= 0:
        value = 0.00001

    roughness_texture = self.metal_roughness_texture
    if roughness_texture:
        slot = material.texture_slots[roughness_texture]
        slot.hardness_factor = value

    material.specular_intensity = 0.04 / (math.pi * pow(value, 4.0))
    material.specular_color = (1.0, 1.0, 1.0)

    self.hardness_float = (2.0 / pow(value, 4.0)) - 2.0
    material.specular_hardness = min(math.floor(self.hardness_float), 511)


def get_texture(self, search_func, index_prop):
    material = self.id_data
    slots = [
        t for t in material.texture_slots
        if t and t.texture and t.texture_coords == 'UV'
    ]

    slot = None
    for slot in slots[::-1]:
        if search_func(slot):
            break
    else:
        return ''

    if (
            bpy.context.space_data
            and bpy.context.space_data.type == 'PROPERTIES'
            and bpy.context.object
    ):
        uv_layers = bpy.context.object.data.uv_layers
        setattr(self, index_prop, uv_layers.find(slot.uv_layer) if slot.uv_layer else 0)

    return slot.texture.name


def _clear_slot_settings(slot):
    slot.use_map_diffuse = False
    slot.use_map_color_diffuse = False
    slot.use_map_alpha = False
    slot.use_map_translucency = False

    slot.use_map_ambient = False
    slot.use_map_emit = False
    slot.use_map_mirror = False
    slot.use_map_raymir = False

    slot.use_map_specular = False
    slot.use_map_color_spec = False
    slot.use_map_hardness = False

    slot.use_map_normal = False
    slot.use_map_warp = False
    slot.use_map_displacement = False

    slot.blend_type = 'MIX'


def set_texture(self, value, current_value, update_func):
    material = self.id_data
    current_index = material.texture_slots.find(current_value)
    slot_index = material.texture_slots.find(value)

    # Clear slot
    if not value:
        if current_index != -1:
            material.texture_slots.clear(current_index)
        return

    # Don't do anything if the correct texture is already set
    if value == current_value:
        return

    bl_texture = bpy.data.textures[value]
    # Texture is not already in a slot on this material
    if current_index == -1 and slot_index == -1:
        slot = material.texture_slots.add()
        slot.texture = bl_texture
        _clear_slot_settings(slot)
        update_func(slot)
        return

    # Adjust existing slot to meet texture criteria
    slot = material.texture_slots[slot_index]
    _clear_slot_settings(slot)
    update_func(slot)
    if slot_index < current_index:
        material.active_texture_index = slot_index
        for _ in range(current_index - slot_index):
            bpy.ops.texture.slot_move(type='DOWN')
            material.active_texture_index -= 1


def get_base_color_texture(self):
    return get_texture(self, lambda t: t.use_map_color_diffuse, 'base_color_text_index')


def set_base_color_texture(self, value):
    def update(slot):
        slot.use_map_color_diffuse = True
        slot.blend_type = 'MULTIPLY'
    set_texture(self, value, get_base_color_texture(self), update)


def get_metal_roughness_texture(self):
    return get_texture(self, lambda t: t.use_map_hardness, 'metal_rough_text_index')


def set_metal_roughness_texture(self, value):
    def update(slot):
        slot.use_map_hardness = True
        slot.hardness_factor = self.roughness_factor
    set_texture(self, value, get_metal_roughness_texture(self), update)


def get_normal_texture(self):
    return get_texture(self, lambda t: t.use_map_normal, 'normal_text_index')


def set_normal_texture(self, value):
    def update(slot):
        slot.use_map_normal = True
    set_texture(self, value, get_normal_texture(self), update)


def get_emissive_texture(self):
    return get_texture(self, lambda t: t.use_map_emit, 'emissive_text_index')


def set_emissive_texture(self, value):
    def update(slot):
        slot.use_map_emit = True
    set_texture(self, value, get_emissive_texture(self), update)


class PbrSettings(bpy.types.PropertyGroup):
    hardness_float = bpy.props.FloatProperty()
    base_color_text_index = 0
    metal_rough_text_index = 0
    normal_text_index = 0
    occlusion_text_index = 0
    emissive_text_index = 0

    base_color_factor = bpy.props.FloatVectorProperty(
        name='Base Color Factor',
        size=4,
        subtype='COLOR',
        min=0.0,
        max=1.0,
        get=get_base_color_factor,
        set=set_base_color_factor,
    )
    base_color_texture = bpy.props.StringProperty(
        name='Texture',
        get=get_base_color_texture,
        set=set_base_color_texture,
    )

    alpha_mode = bpy.props.EnumProperty(
        items=ALPHA_MODE_ITEMS,
        name='Alpha Mode',
        default='OPAQUE',
        get=get_alpha_mode,
        set=set_alpha_mode,
    )

    alpha_cutoff = bpy.props.FloatProperty(
        name='Alpha Cutoff',
        min=0.0,
        max=1.0,
        default=0.5,
    )

    metallic_factor = bpy.props.FloatProperty(
        name='Metallic Factor',
        min=0.0,
        max=1.0,
    )
    roughness_factor = bpy.props.FloatProperty(
        name='Roughness Factor',
        min=0.0,
        max=1.0,
        get=get_roughness_factor,
        set=set_roughness_factor,
    )
    metal_roughness_texture = bpy.props.StringProperty(
        name='Texture',
        get=get_metal_roughness_texture,
        set=set_metal_roughness_texture,
    )

    normal_texture = bpy.props.StringProperty(
        name='Normal',
        get=get_normal_texture,
        set=set_normal_texture,
    )
    occlusion_texture = bpy.props.StringProperty(
        name='Occlusion',
    )

    emissive_factor = bpy.props.FloatVectorProperty(
        name='Emissive Factor',
        size=3,
        subtype='COLOR',
        min=0.0,
        max=1.0,
        get=get_emissive_factor,
        set=set_emissive_factor,
    )
    emissive_texture = bpy.props.StringProperty(
        name='Texture',
        get=get_emissive_texture,
        set=set_emissive_texture,
    )


class PbrExportPanel(bpy.types.Panel):
    bl_idname = 'MATERIAL_PT_pbr_export'
    bl_label = 'PBR Export'
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = 'material'

    @classmethod
    def poll(cls, context):
        return context.material is not None and hasattr(context.material, 'pbr_export_settings')

    def draw(self, context):
        settings = context.material.pbr_export_settings
        self.layout.label('Base Color:')
        box = self.layout.box()
        box.prop(settings, 'base_color_factor', text='Factor')
        box.prop_search(settings, 'base_color_texture', bpy.data, 'textures')
        box.prop(settings, 'alpha_mode')
        box.prop(settings, 'alpha_cutoff')

        self.layout.label('Roughness:')
        box = self.layout.box()
        box.prop(settings, 'metallic_factor', text='Metallic')
        box.prop(settings, 'roughness_factor', text='Factor')
        box.prop_search(settings, 'metal_roughness_texture', bpy.data, 'textures')

        self.layout.label('Emissive:')
        box = self.layout.box()
        box.prop(settings, 'emissive_factor', text='Factor')
        box.prop_search(settings, 'emissive_texture', bpy.data, 'textures')

        self.layout.prop_search(settings, 'normal_texture', bpy.data, 'textures')
        self.layout.prop_search(settings, 'occlusion_texture', bpy.data, 'textures')

        self.layout.prop(context.material.game_settings, 'use_backface_culling')
